from .color import Color, ANSI_8_PALETTE, ANSI_16_PALETTE, ANSI_256_PALETTE
from .format import AnsiFormat
from .markup import AnsiMarkup
from .palette import AnsiPalette
from .string import AnsiString

__version__ = "0.1.0"

palette = AnsiPalette()

__all__ = [
    "ANSI_8_PALETTE", 
    "ANSI_16_PALETTE", 
    "ANSI_256_PALETTE"
    "AnsiFormat",
    "AnsiMarkup",
    "AnsiPalette"
    "AnsiString",
    "Color",
    "palette",
]



from contextlib import contextmanager
from textwrap import indent
from typing import Optional

from zycelium.ansiformat import AnsiFormat


class AnsiMarkup(object):
    def __init__(self, palette=None, term_colors=None):
        self.af = AnsiFormat(palette=palette, term_colors=term_colors)
        self._indent = 0

    def debug(self, text):
        return self.af().gray(text)

    def info(self, text):
        return self.af().blue(text)

    def ok(self, text):
        return self.af().green(text)

    def warning(self, text):
        return self.af().orange(text)

    def error(self, text):
        return self.af().red(text)

    def critical(self, text):
        return self.af().yellow.on.red(text)

    def p(self, text):
        return self.af().black(text)

    def aside(self, text):
        return self.af("[~]").black.on.lightgray.rs_all(" ")(text)

    def note(self, text):
        return self.af("[>]").black.on.lime.rs_all(" ")(text)

    def alert(self, text):
        return self.af("[!]").black.on.orange.rs_all(" ")(text)

    def h1(self, text):
        return self.af("# ").red.b.u(text)

    def h2(self, text):
        return self.af("## ").darkorange.b.u(text)

    def h3(self, text):
        return self.af("### ").blue.b.u(text)

    def h4(self, text):
        return self.af("#### ").black.b.u(text)

    def h5(self, text):
        return self.af("##### ").black.b.u(text)

    def h6(self, text):
        return self.af("###### ").black.b.u(text)

    def li(self, text):
        _text = str(self.af("- ").blue(text))
        _text = indent(_text, " " * self._indent)
        return _text

    @contextmanager
    def indent(self, text=""):
        self._indent = self._indent + 2
        yield
        self._indent = self._indent - 2



from collections import defaultdict
from typing import Optional, Union

import sty

from .color import Color, terminal_colors
from .string import AnsiString


class AnsiFormat(object):
    def __init__(
        self, palette: Optional[list] = None, term_colors: Optional[int] = None
    ):
        self._terminal_colors = (
            term_colors if term_colors is not None else terminal_colors()
        )
        self._custom_palette = palette

    def __call__(self, text=""):
        return AnsiString(text, self._terminal_colors, self._custom_palette)



from collections import defaultdict
from typing import Optional, Union

import sty

from .color import Color, terminal_colors

SHORTCUTS = {
    "b": "ef_bold",
    "bold": "ef_bold",
    "dim": "ef_dim",
    "blink": "ef_blink",
    "hidden": "ef_hidden",
    "inverse": "ef_inverse",
    "strike": "ef_strike",
    "i": "ef_italic",
    "italic": "ef_italic",
    "u": "ef_underl",
    "underline": "ef_underl",
}


class AnsiString(str):
    def __new__(
        cls, text: str, terminal_colors: int = 0, custom_palette: Optional[list] = None
    ):
        return super().__new__(cls, text)

    def __init__(
        self, text: str, terminal_colors: int = 0, custom_palette: Optional[list] = None
    ):
        self._text = text
        self._terminal_colors = terminal_colors
        self._custom_palette = custom_palette
        self._buffer = []
        self._fg_is_set = False

    def __str__(self):
        if self._buffer:
            text = "".join(self._buffer) + self._text
            if self._terminal_colors > 0:
                text += sty.rs.all
            return text
        else:
            return self._text

    def __call__(self, text: str):
        if self._text:
            self._buffer.append(self._text)
        self._text = text
        self._fg_is_set = False
        return self

    def __getattr__(self, name):
        if name.startswith("fg_"):
            return self.fg(name[3:])
        elif name.startswith("bg_"):
            return self.bg(name[3:])
        elif name.startswith("rs_"):
            return self.rs(name[3:])
        elif name.startswith("ef_"):
            return self.ef(name[3:])
        elif name in SHORTCUTS:
            return getattr(self, SHORTCUTS[name])
        elif name == "color":
            return self.color
        else:
            if self._fg_is_set:
                return self.bg(name)
            else:
                return self.fg(name)

    def color(self, name: str):
        if self._fg_is_set:
            return self.bg(name)
        else:
            return self.fg(name)

    def render_color(self, color: Color):
        if self._terminal_colors <= 0:
            return None
        if isinstance(color, str):
            color = Color(color)
        else:
            color = Color(*color)
        if self._terminal_colors == 8:
            return color.ansi8
        elif self._terminal_colors == 16:
            return color.ansi16
        elif self._terminal_colors == 256:
            return color.ansi256
        else:
            if self._custom_palette:
                return color.match_palette(self._custom_palette)
            else:
                return color.RGB

    def fg(self, color: Union[str, list, tuple]):
        if self._terminal_colors == 0:
            return self
        color = self.render_color(color)
        if color is None:
            return self
        self._buffer.append(sty.fg(*color))
        return self

    def bg(self, color: Union[str, list, tuple]):
        if self._terminal_colors == 0:
            return self
        color = self.render_color(color)
        if color is None:
            return self
        self._buffer.append(sty.bg(*color))
        return self

    def rs(self, name: str):
        if self._terminal_colors == 0:
            return self
        ansi = getattr(sty.rs, name)
        self._buffer.append(self._text)
        self._text = ""
        self._buffer.append(ansi)
        return self

    def ef(self, name: str):
        if self._terminal_colors == 0:
            return self
        ansi = getattr(sty.ef, name)
        self._buffer.append(ansi)
        return self

    @property
    def on(self):
        self._fg_is_set = True
        return self

    @property
    def raw(self):
        return repr(str(self))



import logging
import os
import platform
import sys
from math import sqrt
from typing import Optional, Tuple, Union

import sty
from colour import Color as _Color
from colour import COLOR_NAME_TO_RGB

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

IntervalValue = Union[int, float]
RGB255Tuple = Tuple[int, ...]
RGBTuple = Tuple[float, ...]

# Constants

CUBE_INTENSITIES = [
    0x80,
    0x5F,
    0x87,
    0xAF,
    0xD7,
    0xFF,
]

ANSI_8_PALETTE = [
    "#000000",  # black
    "#ff0000",  # red
    "#00ff00",  # lime
    "#ffff00",  # yellow
    "#0000ff",  # blue
    "#ff00ff",  # magenta
    "#00ffff",  # cyan
    "#ffffff",  # white
]

ANSI_16_PALETTE = [
    "#000000",  # black
    "#880000",  # darkred
    "#008800",  # darkgreen
    "#888800",  # orange
    "#000088",  # darkblue
    "#880088",  # darkmagenta
    "#008888",  # darkcyan
    "#888888",  # silver
    "#000000",  # gray
    "#ff0000",  # red
    "#00ff00",  # lime
    "#ffff00",  # yellow
    "#0000ff",  # blue
    "#ff00ff",  # magenta
    "#00ffff",  # cyan
    "#ffffff",  # white
]


def generate_256color(index):
    if index < 16:
        return ANSI_16_PALETTE[index]
    elif index < 232:
        q1, _b = divmod(index - 16, 6)
        q2, _g = divmod(q1, 6)
        _, _r = divmod(q2, 6)
        r = CUBE_INTENSITIES[_r]
        g = CUBE_INTENSITIES[_g]
        b = CUBE_INTENSITIES[_b]
        return f"#{hex(r)[2:]}{hex(g)[2:]}{hex(b)[2:]}"
    else:
        i = 10 * (index - 232) + 8
        h = hex(i)[2:]
        if len(h) == 1:
            h = "0" + h
        return f"#{h}{h}{h}"


ANSI_256_PALETTE = [generate_256color(i) for i in range(256)]


def map_interval(
    from_start: IntervalValue,
    from_end: IntervalValue,
    to_start: IntervalValue,
    to_end: IntervalValue,
    value: IntervalValue,
) -> IntervalValue:
    """
    Map numbers from an interval to another.

    >>> map_interval(0, 1, 0, 255, 0.5)
    127.5

    >>> map_interval(0, 255, 0, 1, 128)  # doctest: +ELLIPSIS
    0.50...

    :param from_start: lower bound of source interval.
    :param from_end: upper bound of source interval.
    :param to_start: lower bound of target interval.
    :param to_end: upper bound of target interval.
    :param value: source value to map to target interval.
    :return: value in target interval.
    """
    return (value - from_start) * (to_end - to_start) / (
        from_end - from_start
    ) + to_start


def rgb_to_RGB255(rgb: RGBTuple) -> RGB255Tuple:
    """
    Convert from Color.rgb's 0-1 range to ANSI RGB (0-255) range.

    >>> rgb_to_RGB255((1, 0.5, 0))
    (255, 128, 0)
    """
    return tuple([int(round(map_interval(0, 1, 0, 255, c))) for c in rgb])


def color_hex_to_int(color):
    return int(color[1:3], 16), int(color[3:5], 16), int(color[5:7], 16)


def distance(color_1, color_2):
    r1, g1, b1 = color_hex_to_int(color_1)
    r2, g2, b2 = color_hex_to_int(color_2)
    return sqrt((r2 - r1) ** 2 + (g2 - g1) ** 2 + (b2 - b1) ** 2)


def match_index(color, palette):
    min_index = -1
    min_distance = 1000000000
    for i, c in enumerate(palette):
        d = distance(color, c)
        if d < min_distance:
            min_index = i
            min_distance = d
    return min_index


def match_color(color, palette):
    index = match_index(color, palette)
    return palette[index]


def terminal_colors(stream=sys.stdout) -> int:
    """
    Get number of supported ANSI colors for a stream.
    Defaults to sys.stdout.

    >>> terminal_colors(sys.stderr)
    0
    """
    colors = 0
    if stream.isatty():
        if platform.system() == "Windows":
            # colorama supports 8 ANSI colors
            # (and dim is same as normal)
            colors = 8
        elif os.environ.get("NO_COLOR", None) is not None:
            colors = 0
        elif os.environ.get("COLORTERM", "").lower() in {"truecolor", "24bit"}:
            colors = 16_777_216
        elif os.environ.get("TERM", "") in {"vt100", "vt200", "vt220"}:
            colors = 0
        elif os.environ.get("TERM", "") in {"xterm"}:
            colors = 8
        elif os.environ.get("TERM", "") in {"xterm-color", "rxvt", "rxvt-88color"}:
            colors = 16
        elif os.environ.get("TERM", "") in {"ansi", "xterm-256color"}:
            colors = 256
        elif os.environ.get("TERM", "").lower() in {"truecolor", "24bit"}:
            colors = 16_777_216
        else:
            # curses is used to autodetect terminal colors on *nix.
            try:
                from curses import setupterm, tigetnum

                setupterm()
                colors = max(0, tigetnum("colors"))
            except ImportError:
                pass
            except:
                pass
    return colors


class Color(_Color):
    @property
    def RGB(self):
        return rgb_to_RGB255(self.rgb)

    @property
    def ansi8(self):
        return (match_index(self.hex_l, ANSI_8_PALETTE),)

    @property
    def ansi16(self):
        return (match_index(self.hex_l, ANSI_16_PALETTE),)

    @property
    def ansi256(self):
        return (match_index(self.hex_l, ANSI_256_PALETTE),)

    def match_palette(self, palette):
        return Color(match_color(self.hex_l, palette)).RGB



from setuptools import setup, find_namespace_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="zycelium.ansiformat",
    version="0.1.0",
    description="ANSI Formatted Text for Terminals.",
    author="Harshad Sharma",
    author_email="harshad@sharma.io",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/zycelium/ansiformat",
    project_urls={
        "Bug Tracker": "https://github.com/zycelium/ansiformat/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Development Status :: 3 - Alpha",
    ],
    packages=['zycelium.ansiformat'],
    package_dir={"": "src"},
    zip_safe=False,
    python_requires=">=3.7",
    install_requires=[
        "sty>=1.0.0rc1",
        "colour>=0.1.5",
    ],
    extras_require={
        "devel": [
            "bump2version>=1.0",
            "pytest>=6.2",
        ],
    },
)
